// This file is part of GElib, a C++/CUDA library for group
// equivariant tensor operations. 
// 
// Copyright (c) 2024, Imre Risi Kondor
//
// This Source Code Form is subject to the terms of the Mozilla
// Public License v. 2.0. If a copy of the MPL was not distributed
// with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


#include "Cnine_base.cpp"
#include "GElib_base.hpp"

#include "SO3CGbank.hpp"
#include "SO3SPHgen.hpp"


namespace GElib{

  SO3CGbank SO3_CGbank;
  SO3SPHgen SO3_SPHgen;

}

